# V1x90.tcl---CAEN V1190/V1290 TDC module control frame

#    code common to V1190 and V1290 

loadsource common.vme

set tdc  0      ;# maxtdc set before  

set alltdcs 0
set allmodules 0

set adjustmode 0


set w2r(GEO)        "GEO_Address"
set w2r(GEOADDRESS) "GEO_Address"
set w2r(INTRLEV)    "Intr_Level"
set w2r(INTRVEC)    "Intr_Vec"
set w2r(MCST)       "MCST|CBLT_Ctrl"
set w2r(EVCOUNTER)  "EvCounter"
set w2r(EVSTORED)   "EvStored"

set w2r(OPT3X0)     "Ctrl.BERR"
set w2r(OPT3X3)     "Ctrl.Event_Mode"

set EGreporting -1

set electronicsvisible 0
set electronicsframe "[self frame-name].electronics"

set optType "none"

set savedirectory CAENtdcParameters


proc Save_Parameters {} {
   global env savedirectory
   global modulename module

   catch {file delete [file join $env(EXPERIMENTBASE) $savedirectory]}
   
   if {![file exists [file join $env(EXPERIMENTBASE) $savedirectory]]} then {
      set z [catch {:file mkdir [file join $env(EXPERIMENTBASE) $savedirectory]} m]
      if {$z != 0} then {midas-warning "Cannot create directory [file join $env(EXPERIMENTBASE) $savedirectory]\n $m"; return}
   }

   set maxmodule [inform vmecontrol inquire [set modulename]count]
   set modulelist {}
   for {set i 1} {$i <= $maxmodule} {incr i} {lappend modulelist $i}

   set module_SAVE $module   ;# currently selected

   set-busy
   set contents ""
   set r 0
   foreach module $modulelist {
         set-footer "Saving settings for module $modulename#$module..."

         click-select-module $module

         set z [catch {open [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module] w} f]
         if {$z != 0} then {
             midas-warning "Cannot create file [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module]\n $f"
             set r 1
             break
         }

#    obtain values for this module

         set vl [do_read_opcode 0x0200]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "AcqMde02 $vl"

         set vl [do_read_opcode 0x1600]   ;# reads 5 words
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "Trig16 $vl"

         set vl [do_read_opcode 0x2300]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "Edge23  $vl"

         set vl [do_read_opcode 0x2600]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "Res26  $vl"

         set vl [do_read_opcode 0x2900]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "Dtime29  $vl"

         set vl [do_read_opcode 0x3200]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "RdOut32  $vl"

         set vl [do_read_opcode 0x3400]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "RdOut34  $vl"

         set vl [do_read_opcode 0x3c00]
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "RdOut3c  $vl"

         set vl [do_read_opcode 0x4500]  ;# reads 8 words
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "Enble45  $vl"

         set vl [do_read_opcode 0x5100]  ;# reads 2 words
         if {$vl == 0x5aa5} then {
             midas-warning "Error reading module parameters\nIf this continues you may need to Reset the module"
             set r 1
             break
         }
         puts $f "Off51  $vl"

         close $f

         append contents "Settings saved for module $modulename#$module to file [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module]\n"
   }
   clear-busy

   catch {close $f}

   if {$r != 0} then {
      midas-warning "Save of module settings to $env(EXPERIMENTDIR)/$savedirectory abandoned"
      clear-footer
      return
   }

   if {$contents == ""} then {
      midas-warning "Warning - Nothing has been saved!"
   } else {
      set-footer "Module settings saved to [file join $env(EXPERIMENTBASE) $savedirectory]"
      midas-report "$contents"
   }

   click-select-module $module_SAVE
}

set Verbose 1
proc Restore_Parameters {} {global Verbose; set Verbose 0; do_Restore_Parameters; set Verbose 1}

proc do_Restore_Parameters {} {
   global env Verbose savedirectory
   global module modulename

   set maxmodule [inform vmecontrol inquire [set modulename]count]
   set modulelist {}
   for {set i 1} {$i <= $maxmodule} {incr i} {lappend modulelist $i}

   set module_SAVE $module   ;# currently selected

   set-busy
   set contents ""
   set r 0
   foreach module $modulelist {
         set-footer "Restoring settings for module $modulename#$module..."

         click-select-module $module

         set z [catch {open [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module] r} f]
         if {$z != 0} then {
             if {$Verbose} then {
                midas-warning "Cannot access file [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module]\n $f"
             } else {
                insert-log "**Warning -  Cannot access file [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module]\n $f"
             }
             set r 1
             break
         }

#    read module settings from file
         set Settings [read $f]
         close $f

#AcqMde02 1 
#Trig16 160 -150 0 0 0 "
#Edge23  2 
#Res26  2 
#Dtime29  0 
#RdOut32  1 
#RdOut34  9 
#RdOut3c  7 
#Enble45  -1 -1 -1 -1 -1 -1 -1 -1 
#Off51  0 0 

         foreach item [split $Settings \n] {
             if {$item == ""} {continue}
             set p [first $item]
             set vl [tail $item]

             switch $p {
                 AcqMde02 {
                     if {[expr $vl & 1] == 1} then {do_AcqMode_opcode OPT00} else {do_AcqMode_opcode OPT01}
                 }

                 Trig16 {
                     do_Trig_opcode OPT10 [first $vl]
                     do_Trig_opcode OPT11 [second $vl]
                     do_Trig_opcode OPT12 [third $vl]
                     do_Trig_opcode OPT13 [fourth $vl]
                     if {[expr [fifth $vl] & 1] == 1} then {do_Trig_opcode OPT14 0} else {do_Trig_opcode OPT15 0}
                 }

                 Edge23 {
                     do_Edge_opcode  OPT22 $vl
                 }

                 Res26 {
                     do_Edge_opcode  OPT24 $vl
                     do_Edge_opcode  OPT25 $vl
                 }

                 Dtime29 {
                     do_Edge_opcode  OPT28 $vl
                 }

                 RdOut32 {
                     if {[expr $vl & 1] == 1} then {do_Readout_opcode OPT30 0} else {do_Readout_opcode OPT31 0}
                 }

                 RdOut34 {
                     do_Readout_opcode OPT33 $vl
                 }

                 RdOut3c {
                     do_Readout_opcode OPT3B $vl
                 }

                 Enble45 {
                     do_set_opcode 0x4400 $vl   ;# write 8 values
                     redisplay_ChEnable
                 }

                 Off51 {
                     do_set_opcode 0x5000 $vl   ;# write 2 values
                     after 500 redisplay_Adjust
                 }

                 Adj53 {
                 }
              }
         }
         append contents "Settings restored for module $modulename#$module from file [file join $env(EXPERIMENTBASE) $savedirectory $modulename#$module]\n"
         
   }

   clear-busy

   if {$r != 0} then {
      if {$Verbose} then {
          midas-warning "Restore of module settings from $env(EXPERIMENTDIR)/$savedirectory abandoned"
      } else {
          insert-log "**Warning  -  Restore of module settings from $env(EXPERIMENTDIR)/$savedirectory abandoned"
      }
      clear-footer
      return
   }

   if {$contents == ""} then {
      if {$Verbose} then {midas-warning "Warning - Nothing has been restored!"}
   } else {
      if {$Verbose} then {
          set-footer "Module settings restored from [file join $env(EXPERIMENTBASE) $savedirectory]"
          midas-report "$contents"
      } else {
          insert-log "$contents"
      }
   }

   click-select-module $module_SAVE

}

proc midas-information {text} {

    set frame [self frame-name].info

    :frame $frame :label  "Information from [self frame-title]" :transient T

    :galley G :deaf T :strings $text :font fixed :rows 50
    :show T
}


proc click-widget {w v} {
    global w2r
    global modulename module
    global allmodules

    set register $w2r($w)
    set value $v

    if {$allmodules == 0} then {
        write-vme-register $modulename#$module.$register $value
    } else {
        write-vme-registers $modulename#\[?,??\].$register $value
    }

    click-redisplay
}

proc update-widget {w} {
    global w2r
    global modulename module

    if {$module < 0} {return}

    set register $w2r($w)
    :modify $w :value [read-vme-register $modulename#$module.$register]
}

proc do_read_register {r} {
    global modulename module
    return [read-vme-register $modulename#$module.$r]
}

proc do_cmd_opcode {OpCode} {
    global modulename module modulelist
    global allmodules
    global EGreport EGreporting
    
    set EGreport_SAVE $EGreport
    set EGreport $EGreporting
    set-busy

    if {$allmodules == 0} then {
        set modules $module
    } else {
        set modules $modulelist
    }
    
    foreach m $modules {
    
      for {set i 50} {$i > 0} {incr i -1} {
        set vl [read-vme-register $modulename#$m.OpCodeCtrl]
        if {$vl == 0x0001} then {
            after 15
            write-vme-register $modulename#$m.OpCode $OpCode
            break
        }
        after 20
      }
      if {$i == 0} then {midas-warning "do_cmd_opcode: timeout for write access: $modulename#$m.OpCodeCtrl = $vl"}
    }
    
    clear-busy
    set EGreport $EGreport_SAVE
}

proc do_set_opcode {OpCode value} {
    global modulename module modulelist
    global maxtdc
    global allmodules
    global EGreport EGreporting
    
    set EGreport_SAVE $EGreport
    set EGreport $EGreporting
    set-busy

    if {$allmodules == 0} then {
        set modules $module
    } else {
        set modules $modulelist
    }
    
    foreach m $modules {
    
      for {set i 50} {$i > 0} {incr i -1} {
        set vl [read-vme-register $modulename#$m.OpCodeCtrl]
        if {$vl == 0x0001} then {
            after 15
            write-vme-register $modulename#$m.OpCode $OpCode
            break
        }
        after 20
      }
      if {$i == 0} then {midas-warning "do_set_opcode: timeout for write access: $modulename#$m.OpCodeCtrl = $vl"}
      
      after 20
      
      set nW 1
      if {$OpCode == 0x4400} {set nW [expr $maxtdc/16]}
      if {$OpCode == 0x4600} {set nW 2}
      if {$OpCode == 0x5000} {set nW 2}
      
      for {set i 50} {$i > 0} {incr i -1} {
        set vl [read-vme-register $modulename#$m.OpCodeCtrl]
        if {$vl == 0x0001} then {
            after 15

            for {set k 0} {$k < $nW} {incr k} {
                write-vme-register $modulename#$m.OpCode [lindex $value $k]
            }
            
            break
        }
        after 20
      }
      if {$i == 0} then {midas-warning "do_set_opcode: timeout for write access: $modulename#$m.OpCodeCtrl = $vl"}
      
    }
    
    clear-busy
    set EGreport $EGreport_SAVE
}

proc do_read_opcode {OpCode} {
    global modulename module
    global maxtdc
    global EGreport EGreporting
    
    set EGreport_SAVE $EGreport
    set EGreport $EGreporting
    set-busy

    set m $module
     
   for {set j 3} {$j > 0} {incr j -1} {
    
      for {set i 50} {$i > 0} {incr i -1} {
        set vl [read-vme-register $modulename#$m.OpCodeCtrl]
        if {$vl == 0x0001} then {
            after 15
            write-vme-register $modulename#$m.OpCode $OpCode
            break
        }
        after 20
      }
      if {$i == 0} then {
      
          clear-busy
          set EGreport $EGreport_SAVE
    
          midas-warning "do_read_opcode: timeout for write access: $modulename#$m.OpCodeCtrl = $vl"
          return 0x5aa5
      }

      after 20
      
      set nR 1
      if {$OpCode == 0x1600} {set nR 5}
      if {$OpCode == 0x4500} {set nR [expr $maxtdc/16]}
      if {$OpCode == 0x4700} {set nR 2}
      if {$OpCode == 0x5100} {set nR 2}
      if {$OpCode == 0x6000} {set nR 2}
      if {$OpCode == 0x7600} {set nR 4}

      for {set i 50} {$i > 0} {incr i -1} {
        set vl [read-vme-register $modulename#$m.OpCodeCtrl]
        if {$vl == 0x0002} then {
            after 15
            
            set vl ""
            for {set k 0} {$k < $nR} {incr k} {
                append vl [format "%hd " [read-vme-register $modulename#$m.OpCode]]
            }

            clear-busy
            set EGreport $EGreport_SAVE
            return $vl
        }
        after 20
      }
      if {$i == 0} then {set-footer "do_read_opcode: timeout for read access: $modulename#$m.OpCodeCtrl = $vl"}
   }
   
    clear-busy
    set EGreport $EGreport_SAVE
    
   set-footer "do_read_opcode: timeout for read access after 3 attempts: $modulename#$m.OpCodeCtrl = $vl"
   
   return 0x5aa5
}

#    Acquisition Mode  -  0x opcodes

#    00   set trigger matching
#    01   set continuous storage
#    02   read acquisition mode
#    03   set keep token
#    04   clear keep token
#    05   load default configuration
#    06   save user configuration
#    07   load user configuration
#    08   set autoload user configuration
#    09   set autoload default configuration

proc do_AcqMode_opcode {w} {

    switch $w {
       OPT00  {do_cmd_opcode 0x0000}
       OPT01  {do_cmd_opcode 0x0100}
       OPT03  {do_cmd_opcode 0x0300}
       OPT04  {do_cmd_opcode 0x0400}
       OPT05  {do_cmd_opcode 0x0500}
       OPT06  {do_cmd_opcode 0x0600}
       OPT07  {do_cmd_opcode 0x0700}
       OPT08  {do_cmd_opcode 0x0800}
       OPT09  {do_cmd_opcode 0x0900}
    }
    redisplay_AcqMode
}

proc  redisplay_AcqMode {} {

     if {![winfo exists [self frame-name].optsel]} {return}
     global optType; if {$optType != "AcqMode"} {return}

    set mode [do_read_opcode 0x0200]
    :modify OPT00 :value $mode
    :modify OPT01 :value [expr ($mode + 1) & 1]
}

#    Trigger setup  -   1x  opcodes

#    10   set window width
#    11   set window offset
#    12   set extra search margin
#    13   set reject margin
#    14   enable subtraction of trigger time
#    15   disable subtraction of trigger time


proc do_Trig_opcode {w v} {

    switch $w {
       OPT10  {do_set_opcode 0x1000 $v}
       OPT11  {do_set_opcode 0x1100 $v}
       OPT12  {do_set_opcode 0x1200 $v}
       OPT13  {do_set_opcode 0x1300 $v}
       OPT14  {do_cmd_opcode 0x1400}
       OPT15  {do_cmd_opcode 0x1500}
    }
    redisplay_Trig
}

proc  redisplay_Trig {} {

     if {![winfo exists [self frame-name].optsel]} {return}
     global optType; if {$optType != "Trig"} {return}
    
    set config [do_read_opcode 0x1600]
    :modify OPT10 :value [first $config]
    :modify OPT11 :value [second $config]
    :modify OPT12 :value [third $config]
    :modify OPT13 :value [fourth $config]

    set mode [fifth $config]
    :modify OPT14 :value $mode
    :modify OPT15 :value [expr ($mode + 1) & 1]

}

#    TDC edge detection and resolution   -    2x opcodes

#    22   set edge detection
#    23   read edge detection
#    24   set LSB of leading/trailing edge
#    25   set pair resolution
#    26   read resolution
#    28   set channel deadtime
#    29   read channel deadtime

proc do_Edge_opcode {w v} {

    switch $w {
       OPT22  {do_set_opcode 0x2200 $v}
       OPT24  {do_set_opcode 0x2400 $v}
       OPT25L  {
           set resolution [do_read_opcode 0x2600]
           do_set_opcode 0x2500 [expr ($resolution & 0xff00) + $v]
       }
       OPT25H  {
           set resolution [do_read_opcode 0x2600]
           do_set_opcode 0x2500 [expr ($resolution & 0x00ff) + ($v << 8)]
       }
       OPT28  {do_set_opcode 0x2800 $v}
    }
    after 500 redisplay_Edge
}

proc redisplay_Edge {} {

     if {![winfo exists [self frame-name].optsel]} {return}
     global optType; if {$optType != "Edge"} {return}

    set mode [do_read_opcode 0x2300]
    :modify OPT22 :value $mode

    if {$mode == 0} then {

#    pair mode

        :modify OPT24 :value -1
        set resolution [do_read_opcode 0x2600]
        :modify OPT25L :value [expr $resolution & 255]
        :modify OPT25H :value [expr ($resolution >> 8) & 255]

    } else {

#    edge mode

        :modify OPT24 :value [do_read_opcode 0x2600]
        :modify OPT25L :value -1
        :modify OPT25H :value -1
    }

    :modify OPT28 :value [do_read_opcode 0x2900]
}

#    TDC Readout  -   3x opcodes

#    30    enable TDC header and EOB in readout
#    31    disable TDC header and EOB in readout
#    32    read status
#    33    set max hits per event
#    34    read max hits per event
#    35    enable TDC error mark
#    36    disable TDC error mark
#    37    enable bypass TDC if error
#    38    disable bypass TDC if error
#    3b    set size of readout FIFO
#    3c    read size 

#    3X0   set control register bit 0
#    3X3   set control register bit 3

proc do_Readout_opcode {w v} {

    switch $w {
       OPT30  {do_cmd_opcode 0x3000}
       OPT31  {do_cmd_opcode 0x3100}
       OPT33  {do_set_opcode 0x3300 $v}
       OPT35  {do_cmd_opcode 0x3500}
       OPT36  {do_cmd_opcode 0x3600}
       OPT37  {do_cmd_opcode 0x3700}
       OPT38  {do_cmd_opcode 0x3800}
       OPT3B  {do_set_opcode 0x3b00 $v}
       OPT3X0 {click-widget $w $v}
       OPT3X3 {click-widget $w $v}
    }
    after 500 redisplay_Readout
}

proc redisplay_Readout {} {
    global modulename module

     if {![winfo exists [self frame-name].optsel]} {return}
     global optType; if {$optType != "Readout"} {return}

    set vl [do_read_opcode 0x3200]
    :modify OPT30 :value $vl
    :modify OPT31 :value [expr ($vl + 1) & 1]

    :modify OPT33 :value [do_read_opcode 0x3400]
    :modify OPT3B :value [do_read_opcode 0x3c00]

    :modify OPT35 :value 0
    :modify OPT36 :value 0
    :modify OPT37 :value 0
    :modify OPT38 :value 0

   set vl [read-vme-register $modulename#$module.Ctrl]

    :modify OPT3X0  :value [expr $vl & 1]
    :modify opt3X3  :value [expr ($vl >> 3) & 1]
}

#    Channel Enable  -  4x opcodes

#    40nn    enable channel nn
#    41nn    disable channel nn
#    42      enable all
#    43      disable all
#    44      write enable pattern
#    45      read enable pattern

proc do_Enable_opcode {w v} {

    switch $w {
       OPT42  {do_cmd_opcode 0x4200}
       OPT43  {do_cmd_opcode 0x4300}
    }

    redisplay_ChEnable
}

proc click-ch-select {i} {

    if {[[:path PATT].bit$i get] == 0} then {
        do_cmd_opcode [expr 0x4000 + $i]
    } else {
        do_cmd_opcode [expr 0x4100 + $i]
    }

    redisplay_ChEnable
}

proc redisplay_ChEnable {} {
    global maxtdc

     if {![winfo exists [self frame-name].optsel]} {return}
     global optType; if {$optType != "ChEnable"} {return}

     set vl [do_read_opcode 0x4500]

     set i 0
     for {set ii 0} {$ii < [expr $maxtdc/16]} {incr ii} {
        set p [lindex $vl $ii]
        for {set jj 0} {$jj < 16} {incr jj} {
           set e [expr ($p >> $jj) & 1]
           [:path PATT].bit$i delete 0 end
           [:path PATT].bit$i insert 0 "$e"
           incr i
        }
     }
}


#    Channel adjust  -  5x opcodes

#    52nn    set channel nn adjust
#    53nn    read channel nn adjust
#    50      set global offset
#    51      read global offset

proc do_Adjust_opcode {w v} {

    switch $w {
       OPT52 {do_adjustment $v}
       OPT50C  {
           set offset [do_read_opcode 0x5100]
           do_set_opcode 0x5000 [list $v [second $offset]]
       }
       OPT50F  {
           set offset [do_read_opcode 0x5100]
           do_set_opcode 0x5000 [list [first $offset] $v]
       }
    }

    after 500 redisplay_Adjust
}

proc redisplay_Adjust {} {

     if {![winfo exists [self frame-name].optsel]} {return}
     global optType; if {$optType != "Adjust"} {return}

     set offset [do_read_opcode 0x5100]
     :modify OPT50C :value [first $offset]
     :modify OPT50F :value [second $offset]
}

proc do_adjustment {v} {
    global adjustmode

    for {set i 0} {$i < 128} {incr i} {
       if {[[:path PATT].bit$i get] == 1 || $adjustmode == 0} then {
          set-footer  "setting ch $i adjust to $v"
          do_set_opcode [expr 0x5200 + $i] $v
       }
    }
    set-footer "All channel adjustments made"
    after 2000 clear-footer
}

proc click-adjust-select {i} {

     set state [[:path PATT].bit$i get]
     set new [expr ($state + 1) & 1]

     [:path PATT].bit$i delete 0 end
     [:path PATT].bit$i insert 0 "$new"
}

proc do_read_buffer {} {
  global CurrentVMEServer modulename module
  global EGreport
  
  set EGreport_Save $EGreport
  set EGreport -1
  set-busy

#    obtain the trigger matching mode

   set mode [do_read_opcode 0x0200]
   set TrgMatch [expr $mode & 1]

   if {$TrgMatch == 0} {set info [do_read_buffer_Cont]} else {set info [do_read_buffer_Event]}

  clear-busy
  set EGreport $EGreport_Save

  midas-information $info
}


proc do_read_buffer_Event {} {
  global modulename module
  

  set info ""
  for {set Offset 0} {$Offset < 2048} {incr Offset 1} {

     set register $modulename#$module.Output_Word
     set data [string trimleft [read-vme-register $register]]

     append info "\nOutput Buffer (FIFO) Offset $Offset\n\n"
     foreach word $data {
         set type [expr $word >> 30 & 3]
         switch $type {
            0 {
             set tdcdatatype [expr $word >> 27 & 7]
             switch $tdcdatatype {
                 1 {set event_type "TDC Header"}
                 3 {set event_type "TDC EOB"}
                 4 {set event_type "TDC Error"}
                 0 {set event_type "TDC Data"}
             }
             if {$tdcdatatype == 0} then {
                if {$modulename == "v1190"} {
                    set channel [expr $word >> 19 & 0x7f]
                    set ed [expr $word >> 26 & 1]
                    set tdc [expr $word  & 0x7ffff]
                    append info "[format "0x%8x  %10s channel= %5i; edge= %3i; time= 0x%5x" $word $event_type $channel $ed $tdc]"
                } else {
                    set channel [expr $word >> 21 & 0x1f]
                    set ed [expr $word >> 26 & 1]
                    set tdc [expr $word  & 0x1fffff]
                    append info "[format "0x%8x  %10s channel= %5i; edge= %3i; time= 0x%6x" $word $event_type $channel $ed $tdc]"
                }
             } else {
                append info "[format "0x%8x  %10s" $word $event_type]"
             }
            }
            1 {
             set event_type "header"
             set geo [expr $word & 0x1f]
             set counter [expr $word >> 5 & 0xffff]
             append info "[format "0x%8x    %8s   module= %5i; event number= %6i" $word $event_type $geo $counter]"
            }
            2 {
             set event_type "trailer"
             set geo [expr $word & 0x1f]
             set status [expr $word >> 20 & 0x7]
             set counter [expr $word >> 5 & 0xffff]
             append info "[format "0x%8x    %8s   module= %5i; status=%2i; event data counter= %8i" $word $event_type $geo $status $counter]"
            }
            3 {
             append info "$word - assumed end of data\n"
             return $info
            }
         }
         append info "\n"
     }
  }
  append info "end of output buffer reached\n"

  return $info
}

proc do_read_buffer_Cont {} {
  global modulename module
  
  set event_type "TDC Data"   ;# continuous mode
  set info ""

  set Offset 0

  for {set Block 0} {$Block < 32} {incr Block 1} {   ;# total of 32 * 64 = 2048 words

     set register $modulename#$module.Output_Buffer
     set data [read-vme-register $register]

     foreach word $data {
         incr Offset 1
         append info "[format "Offset %4d:  " $Offset]"
         set type [expr $word >> 30 & 3]
         switch $type {
            0 {
             set tdcdatatype [expr $word >> 27 & 7]
             if {$tdcdatatype == 0} then {
                if {$modulename == "v1190"} {
                    set channel [expr $word >> 19 & 0x7f]
                    set ed [expr $word >> 26 & 1]
                    set tdc [expr $word  & 0x7ffff]
                    append info "[format "0x%8x  %10s  channel= %5i; edge= %3i; time= 0x%5x" $word $event_type $channel $ed $tdc]"
                } else {
                    set channel [expr $word >> 21 & 0x1f]
                    set ed [expr $word >> 26 & 1]
                    set tdc [expr $word  & 0x1fffff]
                    append info "[format "0x%8x  %10s channel= %5i; edge= %3i; time= 0x%6x" $word $event_type $channel $ed $tdc]"
                }
             } else {
                append info "[format "Word 0x%8x unexpected in Continuous mode" $word]"
             }
            }
            2 -
            1 {
             append info "[format "Word 0x%8x unexpected in Continuous mode" $word]"
            }
            3 {
             append info "$word - assumed end of data\n"
             return $info
            }
         }
         append info "\n"
     }
  }
  append info "end of output buffer reached\n"

  return $info
}


proc click-software {} {
   global modulename
   inform [set modulename]sftwre start
}

proc click-electronics {} {
#  display electronics sub frame
   global electronicsvisible electronicsframe
   if {$electronicsvisible==0} then {make-electronics} else {open-frame $electronicsframe}
   click-redisplay
}

proc click-registers {} {
   global EGreport moduletype modulename module registers_table

   if {[info procs [set moduletype]_registers] == ""} then  \
        {set-footer "No [set moduletype]_registers procedure"; return}

   [set moduletype]_registers

   clear-footer
   set EGreport -1

   set info "[set moduletype] - module $modulename#$module\n\n"

   foreach item $registers_table {
       set description [first $item]
       set register [second $item]
       set data [read-vme-register $modulename#$module.$register]
       append info "[format "%-40s %-20s" $description $data]\n"
   }

  midas-information $info

   set EGreport 0
}

proc click-select-module {v} {
   global debug module maxmodule
   if {$debug} then {puts stderr "click-select-module with argument \{$v\}"}
   set module $v
   if {$module<1} then {set module 1}
   if {$module>$maxmodule} then {set module $maxmodule}
   click-redisplay
}

proc click-all-tdcs {v} {global alltdcs; set alltdcs $v}

proc click-all-modules {v} {global allmodules; set allmodules $v}

proc repaint {} {}

proc redisplay-module {} {
   global modulename module maxmodule allmodules
   global maxtdc

   set maxmodule [inform vmecontrol inquire [set modulename]count]
   set menu {}
   for {set i 1} {$i <= $maxmodule} {incr i} {lappend menu $i}
   :modify MODULE-SEL :strings $menu
   :modify MODULE-NUM :max $maxmodule
   :modify MODULE-NUM :value $module
   :modify ALL-MODULES :value $allmodules

   update-widget GEO
   update-widget EVCOUNTER
   update-widget EVSTORED

   set vers [read-vme-register $modulename#$module.Version]
   if {$modulename == "v1190"} {
       if {[expr $vers & 255] == 0} {:modify VERSION :value "A"; set maxtdc 128} else {:modify VERSION :value "B"; set maxtdc 64}
   } else {
       if {[expr $vers & 255] == 0} {:modify VERSION :value "A"; set maxtdc 32} else {:modify VERSION :value "N"; set maxtdc 16}
   }

}

proc redisplay-electronics {} {
   global electronicsvisible
   global modulename module w2r

   if {$electronicsvisible==0} then {return}

   foreach w {GEOADDRESS INTRLEV INTRVEC} {update-widget $w}

#    Status register

   set value [read-vme-register $modulename#$module.Status]

   :modify S0    :value [expr $value & 1]
   :modify S1    :value [expr $value >> 1 & 1]
   :modify S2    :value [expr $value >> 2 & 1]
   :modify S3    :value [expr $value >> 3 & 1]
   :modify S4    :value [expr $value >> 4 & 1]
   :modify S5    :value [expr $value >> 5 & 1]
   :modify S6    :value [expr $value >> 6 & 15]
   :modify S10   :value [expr $value >> 10 & 1]
   :modify S12   :value [expr $value >> 12 & 3]
   :modify S14   :value [expr $value >> 14 & 1]
   :modify S15   :value [expr $value >> 15 & 1]
}

proc reset-module {} {
    global modulename module allmodules

    if {$allmodules == 0} then {
        write-vme-register $modulename#$module.Reset 0
    } else {
        write-vme-registers $modulename#\[?,??\].Reset 0
    }
    
    click-redisplay
}

proc clear-module {} {
    global modulename module allmodules

    if {$allmodules == 0} then {
        write-vme-register $modulename#$module.Clear 0
    } else {
        write-vme-registers $modulename#\[?,??\].Clear 0
    }
    
    click-redisplay
}

proc click-redisplay {} {
   global EGreport

   clear-footer
   set EGreport -1
   set-busy

   redisplay-module
   redisplay-electronics

   clear-busy
   set EGreport 0
}


proc click arg {
   global debug expertlist adjustmode
   if {$debug} then {insert-debug "click with argument \{$arg\}"}
   set f [lindex $arg 0]
   set c [lindex $arg 1]
   set w [lindex $arg 2]
   set v [lindex $arg 3]
   set i [lindex $arg 4]

   if {$c==":quit"} then {click-quit; return}
   clear-footer
   switch -glob $w {

      MODULE-SEL      -
      MODULE-NUM  {click-select-module $v}

      ALL-MODULES {click-all-modules $v}
      ALL-TDCS    {click-all-tdcs $v}

      OPT0*      {do_AcqMode_opcode $w}
      OPT1*      {do_Trig_opcode $w $v}
      OPT2*      {do_Edge_opcode $w $v}
      OPT3*      {do_Readout_opcode $w $v}
      OPT4*      {do_Enable_opcode $w $v}
      OPT5*      {do_Adjust_opcode $w $v}

      ADJSET     {set adjustmode $v}

      ENGINEERING       {click-$v}
      REDISPLAY  	{click-redisplay}

      EXPERT  {
         set action [lindex $expertlist $i]
         set r [midas-confirmation "You are about to execute the action \"[first $action]\"\n\nAre you sure?"]
         if {$r == 1} then {eval [second $action]}
      }

      default           {
           set p "click_[lowercase $w]"
           if {[info procs $p] == "$p"} then {$p [list $v $i]} else {insert-debug "$arg"}
      }
   }
}


proc make-frame {} {
  global env expertlist
  global module maxmodule modulelist

  :frame [self frame-name] :label "[self frame-title]" :show-footer T \
      :icon-image "@$env(ICONHOME)/vme_module" \
      :icon-name "[self icon-title]" :resize F

   :panel MODULE  :layout T :padx 10 :pady 10
     :layout H; :row-gap 10

     :number MODULE-NUM  :label "Module number:" :min 1 :max $maxmodule :w 5 :value $module
     :menubutton MODULE-SEL :label "select:" :strings $modulelist

     :col-gap 10
     :text EVCOUNTER :label "Trigger Counter" :w 8 :deaf T
     :text EVSTORED  :label "Events Stored" :w 8 :deaf T
     :checkbox ALL-MODULES :label "Act on ALL modules?"

     :next-row T
     :button OPT42 :label "Enable all channels"
     :button OPT43 :label "Disable all channels"
     :button CHENBLE :label "Channel Enable Options"   :command select_ChEnable_options
     :text VERSION :label "Version" :w 2 :deaf T
     :text GEO :label "GEO Address" :w 6 :deaf T
  
     :next-row T
     :button ACQMODE :label "Acquisition Mode Options" :command select_acqmode_options
     :button TRIG    :label "Trigger Options"          :command select_trigger_options
     :button EDGE    :label "Edge Detection Options"   :command select_edge_detection_options
     :button READOUT :label "Readout Options"          :command select_readout_options

     :button ADJUST  :label "Channel Adjust"           :command select_adjust_options

   :panel CLICK :layout B :padx 10 :after FOOTER
     :layout H; :col-gap 30; :row-gap 10
      :vskip 10
      :button REDISPLAY
      :button SAVE    :label "Save Setup"    :command Save_Parameters
      :button RESTORE :label "Restore Setup" :command do_Restore_Parameters
      :menubutton ENGINEERING :label "Engineering"  :menu [list electronics software registers]
      :button MEMRD   :label "read output buffer" :command do_read_buffer

#      set menu {}
#      foreach item $expertlist {lappend menu [first $item]}
#      :menubutton EXPERT :label "Save/Restore Options"  :strings $menu

  :show T

}

proc select_acqmode_options {} {
    global moduletype

  :frame "[self frame-name].optsel" :label "$moduletype Acquisition Mode Options" :resize F :transient T
     global optType; set optType "AcqMode"

   :panel P :layout T :padx 5 :pady 5
     :layout V; :row-gap 5

     :checkbox OPT00   :label "Set Trigger Matching"
     :checkbox OPT01   :label "Set Continuous Storage"

     :next-row T
     :button OPT03   :label "Set keep token"
     :button OPT04   :label "Clear keep token"

     :next-row T
     :button OPT05   :label "Load default configuration"
     :button OPT06   :label "Save user configuration"
     :button OPT07   :label "Load user configuration"

     :next-row T
     :button OPT08   :label "Set auto load user configuration"
     :button OPT09   :label "set auto load default configuration"

     :next-row T
     :button UPDATEOPT0x   :label "Update"   :command redisplay_AcqMode

  :show T

  redisplay_AcqMode
}

proc select_trigger_options {} {
    global moduletype

 set valuex 24c

  :frame "[self frame-name].optsel" :label "$moduletype Triggering Options" :resize F :transient T
     global optType; set optType "Trig"

   :panel P :layout T :padx 5 :pady 5
     :layout V; :row-gap 5

     :slider OPT10   :label "Window Width (25ns)"  :min 1 :max 4095 :valuex $valuex
     :slider OPT11   :label "Window Offset (25ns)" :min -2048 :max +40 :valuex $valuex
     :slider OPT12   :label "Extra Search Margin (25ns)" :min 0 :max 4095 :valuex $valuex
     :slider OPT13   :label "Reject Margin (25ns)" :min 0 :max 4095 :valuex $valuex
     :checkbox OPT14 :label "Enable Subtraction of Trigger Time"  :valuex 32c
     :checkbox OPT15 :label "Disable Subtraction of Trigger Time" :valuex 32c

     :button UPDATEOPT3x   :label "Update"   :command redisplay_Trig

  :show T

  redisplay_Trig
}

proc select_edge_detection_options {} {
    global moduletype

 set valuex 40c

  :frame "[self frame-name].optsel" :label "$moduletype Edge Detection Options" :resize F :transient T
     global optType; set optType "Edge"

   :panel P :layout T :padx 5 :pady 5
     :layout V; :row-gap 5

     :choice OPT22    :label "Edge Detection Configuration" :valuex $valuex \
       :strings [list "pair mode" "only leading" "only trailing" "trailing and leading"]
     if {$moduletype == "CAEN V1190"} {
         :choice OPT24    :label "Set LSB of leading/trailing edge" :valuex $valuex \
           :strings [list "800 ps" "200 ps" "100 ps"]
     } else {
         :choice OPT24    :label "Set LSB of leading/trailing edge" :valuex $valuex \
           :strings [list "800 ps" "200 ps" "100 ps" "25ps"]
     }
     :choice OPT25L   :label "Resolution of leading edge arrival time" :valuex $valuex \
       :strings [list "100 ps" "200 ps" "800 ps" "1.6 ns" "3.12 ns" "6.25 ns" "12.5 ns"]
     :choice OPT25H   :label "Resolution of pulse width (pair detection)" :valuex $valuex \
       :strings [list "100ps" "200ps" "800ps" "1.6ns" "3.12ns" "6.25ns" "12.5ns" "25ns" "50ns" "100ns" 200ns 400ns 800ns]
     :choice OPT28    :label "Channel dead time" :valuex $valuex \
       :strings [list "5 ns" "10 ns" "30 ns" "100 ns"]

     :button UPDATEOPT2x   :label "Update"   :command redisplay_Edge

  :show T

  redisplay_Edge
}

proc select_readout_options {} {
    global moduletype

 set valuex 32c

  :frame "[self frame-name].optsel" :label "$moduletype Readout Options" :resize F :transient T
     global optType; set optType "Readout"

   :panel P :layout T :padx 5 :pady 5
     :layout V; :row-gap 5

     :checkbox OPT30    :label "Enable TDC Header and EOB" :valuex $valuex
     :checkbox OPT31    :label "Disable TDC Header and EOB" :valuex $valuex
     :choice OPT33      :label "Maximum hits per event" :valuex $valuex   \
        :strings [list 0 1 2 4 8 16 32 64 128 "no limit"]
     :checkbox OPT35    :label "Enable TDC error mark" :valuex $valuex
     :checkbox OPT36    :label "Disable TDC error mark" :valuex $valuex
     :checkbox OPT37    :label "Enable bypass TDC if error" :valuex $valuex
     :checkbox OPT38    :label "Disable bypass TDC if error" :valuex $valuex
     :choice OPT3B      :label "Set readout FIFO size" :valuex $valuex   \
        :strings [list 32 64 96 128 160 192 224 256]
     :checkbox OPT3X0     :label "BERR enable"  :valuex $valuex
     :checkbox OPT3X3     :label "Write Empty Event"  :valuex $valuex

     :button UPDATEOPT3x   :label "Update"   :command redisplay_Readout

  :show T

  redisplay_Readout
}

proc select_ChEnable_options {} {
    global env moduletype maxtdc

  :frame "[self frame-name].optsel" :label "$moduletype Channel Enable" :resize F :transient T
     global optType; set optType "ChEnable"

   :panel P :layout T :padx 5 :pady 5
     :layout V; :row-gap 5

        :canvas PATT  :w 400 :h 270
        [:path PATT] config -background $env(tkframebackground)

        set x 20
        set y 20
        set i 0
        for {set ii 0} {$ii < [expr $maxtdc/16]} {incr ii} {

           :draw text :value $x $y -text "$i"
           incr x 30

           for {set jj 0} {$jj < 16} {incr jj} {
              entry [:path PATT].bit$i -relief flat -width 1 -justify right
              [:path PATT].bit$i insert 0 "0"
              :draw window :tag PATTbit$i :value $x $y -window [:path PATT].bit$i 
              bind [:path PATT].bit$i <ButtonPress-1>  "click-ch-select $i"
              incr x 20
              incr i
           }

           incr x 10

           :draw text :value $x $y -text "[expr $i-1]"

           incr y 32
           set x 20
        }

     :button UPDATEOPT4x   :label "Update"   :command redisplay_ChEnable

  :show T

   redisplay_ChEnable
}

proc select_adjust_options {} {
    global env adjustmode moduletype

 set valuex 18c

  :frame "[self frame-name].optsel" :label "$moduletype Channel Adjust Options" :resize F :transient T
     global optType; set optType "Adjust"

   :panel P1 :layout T :padx 5 :pady 5
     :layout V; :col-gap 20; ; :row-gap 10

     :message OFFSET :label "Global Offset" :bold T
     :slider OPT50F  :label "fine"   :min 0 :max 31   :valuex $valuex
     :slider OPT50C  :label "coarse" :min 0 :max 2047 :valuex $valuex

     :message ADJUST  :label "Channel adjustment" :bold T
     :slider OPT52  :label ""  :min 0 :max 127  :valuex $valuex
     :choice ADJSET :label "Apply Channel Adjust To" :strings [list All Selected] :value $adjustmode

   :panel P2 :layout T :padx 5 :pady 5
     :layout V; :row-gap 5

        :canvas PATT  :w 400 :h 270
        [:path PATT] config -background $env(tkframebackground)

        set x 20
        set y 20
        set i 0
        for {set ii 0} {$ii < 8} {incr ii} {

           :draw text :value $x $y -text "$i"
           incr x 30

           for {set jj 0} {$jj < 16} {incr jj} {
              entry [:path PATT].bit$i -relief flat -width 1 -justify right
              [:path PATT].bit$i insert 0 "0"
              :draw window :tag PATTbit$i :value $x $y -window [:path PATT].bit$i 
              bind [:path PATT].bit$i <ButtonPress-1>  "click-adjust-select $i"
              incr x 20
              incr i
           }

           incr x 10

           :draw text :value $x $y -text "[expr $i-1]"

           incr y 32
           set x 20
        }

   :panel P3 :layout B :padx 5 :pady 5 :after FOOTER
     :layout H; :col-gap 20

     :button UPDATEOPT5x   :label "Update"   :command redisplay_Adjust

  :show T

  redisplay_Adjust
}

proc make-electronics {} {
  global env electronicsframe electronicsvisible moduletype

 set valuex 32c

  :frame "$electronicsframe" :label "$moduletype TDC Electronics" :resize F :transient T

   :panel P20 :layout T :padx 5 :pady 5
     :layout V; :row-gap 5; :col-gap 10

      :text GEOADDRESS :label "GEO Address" :w 8 :valuex $valuex  :deaf T
      :text INTRLEV    :label "Interrupt level" :w 8 :valuex $valuex  :deaf T
      :text INTRVEC    :label "Interrupt vector" :w 8 :valuex $valuex  :deaf T

      :checkbox S0     :label "Output Buffer has data" :valuex $valuex :deaf T
      :checkbox S1     :label "Output Buffer is Almost Full" :valuex $valuex :deaf T
      :checkbox S2     :label "Output Buffer is full" :valuex $valuex :deaf T
      :choice S3       :label "Trigger Match mode" :strings [list "continuous storage" "trigger matching"] \
            :valuex $valuex :deaf T

      :choice S4       :label "TDC Header and EOB" :strings [list disabled enabled] :valuex $valuex :deaf T
      :choice S5       :label "Termination" :strings [list OFF ON] :valuex $valuex :deaf T 
      :checkbox S6     :label "TDC error" :valuex $valuex :deaf T 
      :checkbox S10    :label "BERR" :valuex $valuex :deaf T
      :choice S12      :label "Resolution" :strings [list 800ps 200ps 100ps 25ps] :valuex $valuex :deaf T
      :checkbox S14    :label "Pair mode" :valuex $valuex :deaf T
      :checkbox S15    :label "Trigger lost" :valuex $valuex :deaf T

     :layout H; :next-row T

     :button RESET    :label "Module Reset"         :command reset-module
     :button CLEAR    :label "Module clear"         :command clear-module

     :next-row T
     :button UPDATE   :label "Electronics Update"   :command redisplay-electronics
     :button SHOWREGS :label "Show register values" :command click-registers

  :show T
 set electronicsvisible 1
}
