# vmediag.tcl---VME Diagnostics frame


set debug 0
set started 0

set frame VMEDIAG
set frametitle "VME Diagnostics"
set icontitle  "VMEDiag"

if {[info exists env(VMEServer)] == 0} then {set resource vme}  \
   else {set resource $env(VMEServer)}

set mchoice {"0x09" "0x0a" "0x0d" "0x0e" "0x39" "0x3a" "0x3d" "0x3e" "0x29" "0x2d"}
set wchoice {16 24 32}
set dchoice {8 16 32}
set modechoice {"non-privileged" "supervisory"}
set modeopts {n s}
set uchoice {"data" "program" }
set uopts {d p}

set pchoice {"rw" "ro" "wo"}
set lchoice {"0" "1" "2" "3" "4" "5" "6" "7" "8" "9" "10" "11" "12" "13" "14" "15" "16"}
set bchoice {"0" "1" "2" "3" "4" "5" "6" "7" "8" "9" "10" "11" "12" "13" "14" "15" "16"}
set zchoice {"x" "d" "b"}
set schoice {"0" "1"}

set classchoice {"xVME" "bVME"}

set debugchoice {"0" "0x01" "0x02" "0x04" "0x08" "0x10" "-1"}

set vmedata "0x0000"
set regdata "0x0000"
set vmesdata "00"
set textdata ""


proc make-frame {} {
  global env
  global names
  global achoice mchoice dchoice
  global modechoice wchoice uchoice pchoice lchoice bchoice
  global zchoice schoice classchoice debugchoice
  global textdata vmedata vmesdata regdata

  :frame [self frame-name] :label "[self frame-title]" :show-footer T \
      :icon-image "@$env(ICONHOME)/default" \
      :icon-name "[self icon-title]"  :resize F
   
   :panel DIAG-PANEL :padx 10
      :layout H; :row-gap 5; :col-gap 5
      :vskip 10

      :message M1 :label "VME.Address" :bold T
      :next-row T
      :text DIAG-A :label "Addr" :w 12 :value "0x00000000"
      :choice-stack DIAG-M :label "Addr Modifier" :exclusive T :strings $mchoice :value 0
      :choice-stack DIAG-D :label "Data Width" :exclusive T :strings $dchoice :value 2

      :next-row T
      :button ADDR-W :label "Write VME.Address"
      :button ADDR-I :label "Initialise VME.Address"
      :next-row T
      :text ADDR-RD :label "" :w 30 :value "$textdata" :deaf T
      :button ADDR-R :label "Read VME.Address"
      :row-gap 15
      :next-row T
      :message M2 :label "VME.Execute" :bold T
      :row-gap 5
      :next-row T
      :text EXEC-WD :label "" :w 10 :value "$vmedata"
      :button EXEC-W :label "Write VME.Execute"
      :next-row T
      :text EXEC-RD :label "" :w 10 :value "$vmedata" :deaf T
      :button EXEC-R :label "Read VME.Execute"
      :row-gap 10
      :next-row T
      :message M3 :label "VME.Status" :bold T
      :row-gap 5
      :next-row T
      :text STATUS-RD :label "" :w 3 :value "$vmesdata" :deaf T
      :button STATUS-R :label "Read VME.Status"
      :col-gap 100
      :choice-stack DEBUG :label "VME.Debug" :exclusive T :strings $debugchoice :value 0
      :row-gap 10; :col-gap 5
      :next-row T
      :message M4 :label "VME.Data" :bold T
      :row-gap 5
      :next-row T
      :text DATA-RD :label "" :w 10 :value "$vmedata" :deaf T
      :button DATA-R :label "Read VME.Data"
      :row-gap 30
      :next-row T
      :text REG-NAME :label "Register Name:" :w 50 :value "$textdata"
      :row-gap 10
      :next-row T
      :choice-stack REG-CLASS :label "Class" :exclusive T :strings $classchoice :value 0
      :next-row T
      :message M5 :label "Register Attributes"  :bold T
      :row-gap 5; :col-gap 10
      :next-row T
      :text REG-BA :label "Base Addr" :w 12 :value "0x00000000"
      :text REG-O :label "Addr Offset" :w 12 :value "0x00000000"
      :next-row T
      :choice-stack REG-W :label "Addr Width" :exclusive T :strings $wchoice :value 2
      :choice-stack REG-D :label "Data Width" :exclusive T :strings $dchoice :value 1
      :next-row T
      :choice-stack REG-U :label Use :exclusive T :strings $uchoice :value 0
      :choice-stack REG-MODE :label Mode :exclusive T :strings $modechoice :value 0
      :choice-stack REG-AMOD :label "Addr Modifier" :exclusive T :strings $mchoice :value 0
      :next-row T
      :choice-stack REG-P :label p :exclusive T :strings $pchoice :value 0
      :choice-stack REG-L :label l :exclusive T :strings $lchoice :value 0
      :choice-stack REG-B :label b :exclusive T :strings $bchoice :value 0
      :choice-stack REG-Z :label z :exclusive T :strings $zchoice :value 0
      :choice-stack REG-S :label s :exclusive T :strings $schoice :value 0
      :row-gap 10; :col-gap 5
      :next-row T
      :button REG-CR :label "Create Register"
      :button REG-WA :label "Write Register Attributes"
      :button REG-I  :label "Initialise Register"
      :next-row T
      :text REG-RAD :label "" :w 65 :value "$textdata" :deaf T
      :next-row T
      :button REG-RA :label "Read Register Attributes"
      :next-row T
      :text REG-WDATA :label "" :w 10 :value "$regdata"
      :button REG-WD :label "Write Register"
      :col-gap 10
      :text REG-RDATA :label "" :w 10 :value "$textdata" :deaf T
      :button REG-RD :label "Read Register"
   
    :show T
}

set diag_a 0x00000000
set diag_m 0x09
set diag_d 32

set reg_name ""
set reg_class "xVME"

set reg_a 0x00000000
set reg_o 0x00000000
set reg_w 32
set reg_d 32
set reg_u d
set reg_mode "n"
set reg_amod "0x09"
set reg_p rw
set reg_l 0
set reg_b 0
set reg_z d
set reg_s 0

#proc EG {arg} {puts stdout "$arg"; return 16384}

proc write-vme-register {r v} {
   global resource
   EG  "register :name '/$resource/VME.$r' :value '$v'"
}

proc read-vme-register {r} {
   global resource
   EG  "register :name '/$resource/VME.$r' ?value"
}

proc write-register {r v} {
   global resource
   EG  "register :name '/$resource/$r' :value '$v'"
}

proc read-register {r} {
   global resource
   EG "register :name '/$resource/$r' ?value"
}

proc write-attributes {r v} {
   global resource
   EG  "register :name '/$resource/$r' :object '$v'"
}

proc read-attributes {r} {
   global resource
   EG "register :name '/$resource/$r' ?object"
}

proc initialise-register {r} {
   global resource
   EG  "register :name '/$resource/$r' !initialise"
}

proc click-redisplay {} {
   global env server
   :modify CA-REDISPLAY :busy T
   read-card $server
   :modify CA-REDISPLAY :busy F
}

proc click-write-attributes {} {
   global reg_class reg_name
   global reg_a reg_o reg_w reg_d reg_u reg_m reg_p reg_l reg_b reg_z reg_s
   global reg_mode reg_amod

    switch $reg_class {
       xVME  {write-attributes "$reg_name"  \
     "-a$reg_a -o$reg_o -w$reg_w -d$reg_d -u$reg_u -m$reg_mode -p$reg_p \
          -l$reg_l -b$reg_b -z$reg_z -s$reg_s"}
       bVME  {write-attributes "$reg_name"  \
     "-a$reg_a -o$reg_o -w$reg_w -d$reg_d -u$reg_u -m$reg_mode -p$reg_p \
          -l$reg_l -s$reg_s"}
    }
}

proc click-addr-redisplay {} {
  :modify ADDR-RD :value [read-vme-register "Address"]
}

proc click-exec-redisplay {} {
  :modify EXEC-RD :value [read-vme-register "Execute"]
}

proc click-status-redisplay {} {
  :modify STATUS-RD :value [read-vme-register "Status"]
}

proc click-data-redisplay {} {
  :modify DATA-RD :value [read-vme-register "Data"]
}

proc click-reg-redisplay {} {
  global reg_name
  :modify REG-RDATA :value "[read-register "$reg_name"]"
}

proc click-rega-redisplay {} {
  global reg_name
  :modify REG-RAD :value "[read-attributes "$reg_name"]"
}

proc click arg {
   global env debug resource diag_a diag_m diag_d
   global reg_name reg_class
   global reg_a reg_o reg_w reg_d reg_u reg_mode reg_amod reg_p reg_l reg_b reg_z reg_s
   global vmedata regdata
   global mchoice wchoice dchoice pchoice lchoice bchoice zchoice schoice
   global modechoice modeopts uchoice uopts
   global debugchoice classchoice
   if {$debug} then {puts stderr "click with argument \{$arg\}"}
   set f [lindex $arg 0]
   set c [lindex $arg 1]
   set w [lindex $arg 2]
   set v [lindex $arg 3]
   if {$f=="ER"} then {remove-eg-error-report self; return}
   if {$c==":quit"} then {click-quit; return}
   switch $w {

	DIAG-A	{set diag_a $v}
	DIAG-M	{set diag_m [lindex $mchoice $v]}
	DIAG-D	{set diag_d [lindex $dchoice $v]}

	ADDR-W	{write-vme-register "Address" "-a $diag_a -m $diag_m -d $diag_d"}
	ADDR-I	{eg "(register :name '/$resource/VME.Address' !initialise)"}
	ADDR-R	{click-addr-redisplay}

	EXEC-WD	{set vmedata $v}
	EXEC-W	{write-vme-register "Execute" "$vmedata"}
	EXEC-R	{click-exec-redisplay}

	STATUS-R {click-status-redisplay}

	DATA-R	{click-data-redisplay}

	REG-NAME	{set reg_name $v}
	REG-CLASS	{set reg_class [lindex $classchoice $v]}

	REG-BA	{set reg_a $v}
	REG-O	{set reg_o $v}
	REG-W	{set reg_w [lindex $wchoice $v]}
	REG-D	{set reg_d [lindex $dchoice $v]}
	REG-U	{set reg_u [lindex $uopts $v]}
	REG-MODE {set reg_mode [lindex $modeopts $v]}
	REG-AMOD {set reg_amod [lindex $mchoice $v]}
	REG-P	{set reg_p [lindex $pchoice $v]}
	REG-L	{set reg_l [lindex $lchoice $v]}
	REG-B	{set reg_b [lindex $bchoice $v]}
        REG-Z   {set reg_z [lindex $zchoice $v]}
	REG-S	{set reg_s [lindex $schoice $v]}

	REG-CR	{EG "register :name '/$resource/$reg_name' :class '$reg_class' !new"}
	REG-I	{initialise-register "$reg_name"}
	REG-WA	{click-write-attributes}
	REG-RA  {click-rega-redisplay}

	REG-WDATA {set regdata $v}
	REG-WD	{write-register "$reg_name" "$regdata"}
	REG-RD	{click-reg-redisplay}

	DEBUG	{write-vme-register Debug [lindex $debugchoice $v]}

      default    {puts stdout $arg}
      }
   }


proc start args {
   global started
   if {$started == 0} then {make-frame} else {open-frame}
   set started 1
}

proc click-quit {} {close-frame}

proc finish args {destroy-frame}


